@@ -3,7 +3,7 @@ import uuid
 from jsonpath_ng.jsonpath import Fields, Slice, Where
 from jsonpath_ng.ext import parse as parse_ext
 
-from .model.cohort_data_entity import CohortData, FieldEntry
+from .model.cohort_data_entity import CohortData, CohortData
 from .model.resource_definition_entity import ResourceDefinition
 from .model.resource_link_entity import ResourceLink
 from . import fhir_formatting
@@ -27,11 +27,31 @@ def create_transaction_bundle(resource_definition_entities: List[ResourceDefinit
         add_resource_to_transaction_bundle(root_bundle, fhir_resource)
     return root_bundle
 
+def create_singular_resource(singleton_entity_name: str, resource_definition_entities: List[ResourceDefinition], resource_link_entities: List[ResourceLink], cohort_data: CohortData, index = 0):
+    created_resources = {}
+    for resource_definition in resource_definition_entities:
+        entity_name = resource_definition.entity_name
+        #Create and collect fhir resources
+        fhir_resource = create_fhir_resource(resource_definition, cohort_data, index)
+        created_resources[entity_name] = fhir_resource
+        if entity_name == singleton_entity_name:
+            singleton_fhir_resource = fhir_resource
+    add_default_resource_links(created_resources, resource_link_entities)
+    create_resource_links(created_resources, resource_link_entities, preview_mode=True)
+    return singleton_fhir_resource
+
 #Initialize root bundle definition
 def initialize_bundle():
     root_bundle = {}
     root_bundle['resourceType'] = 'Bundle'
     root_bundle['id'] = str(uuid.uuid4())
+    root_bundle['meta'] = {
+        'security': [{
+            'system': 'http://terminology.hl7.org/CodeSystem/v3-ActReason',
+            'code': 'HTEST',
+            'display': 'test health data'
+        }]
+    }
     root_bundle['type'] = 'transaction'
     root_bundle['entry'] = []
     return root_bundle
@@ -43,24 +63,39 @@ def initialize_resource(resource_definition):
     initial_resource['id'] = str(uuid.uuid4()).strip()
     if resource_definition.profiles:
         initial_resource['meta'] = {
-            'profile': resource_definition.profiles
+            'profile': resource_definition.profiles,
+            'security': [{
+                'system': 'http://terminology.hl7.org/CodeSystem/v3-ActReason',
+                'code': 'HTEST',
+                'display': 'test health data'
+            }]
         }
     return initial_resource
 
 # Creates a fhir-json structure from a resource definition entity and the patient_data_sheet
 def create_fhir_resource(resource_definition: ResourceDefinition, cohort_data: CohortData, index = 0):
     resource_dict = initialize_resource(resource_definition)
-    #Get field entries for this entitiy
-    try:
-        all_field_entries = cohort_data.entities[resource_definition.entity_name].fields
-    except KeyError:
+    #Get field entries for this entity
+    header_entries_for_resourcename = [
+        headerEntry
+        for headerEntry in cohort_data.headers
+        if headerEntry.entityName == resource_definition.entity_name
+    ]
+    dataelements_for_resourcename = {
+        field_name: value
+        for (entity_name, field_name), value in cohort_data.patients[index].entries.items()
+        if entity_name == resource_definition.entity_name
+    }
+    if len(dataelements_for_resourcename.keys()) == 0:
         print(f"WARNING: Patient index {index} - Create Fhir Resource Error - {resource_definition.entity_name} - No columns for entity '{resource_definition.entity_name}' found for resource in 'PatientData' sheet")
         return resource_dict
+        all_field_entries = cohort_data.entities[resource_definition.entity_name].fields
     #For each field within the entity
-    for field_entry_key, field_entry in all_field_entries.items():
-        #Create a jsonpath from each provided json path and value for this resource
-        if field_entry.values and len(field_entry.values) > index:
-            create_structure_from_jsonpath(resource_dict, field_entry.jsonpath, resource_definition, field_entry, field_entry.value_type, field_entry.values[index])
+    for fieldname, value in dataelements_for_resourcename.items():
+        header_element = next((header for header in header_entries_for_resourcename if header.fieldName == fieldname), None)
+        if header_element is None:
+            print(f"WARNING: Field Name {fieldname} - No Header Entry found.")
+        create_structure_from_jsonpath(resource_dict, header_element.jsonpath, resource_definition, header_element.value_type, value)
     return resource_dict
 
 #Create a resource_link for default references in the cases where only 1 resourceType of the source and destination exist
@@ -128,12 +163,21 @@ def add_default_resource_links(created_resources: dict, resource_link_entities:
     return
         
             
-#Create resource references/links with created entities
-def create_resource_links(created_resources, resource_link_entites):
+#List function to create resource references/links with created entities
+def create_resource_links(created_resources, resource_link_entites, preview_mode = False):
+    #TODO: Build resource links
+    print("Building resource links")
+    for resource_link_entity in resource_link_entites:
+        create_resource_link(created_resources, resource_link_entity, preview_mode)
+    return
+    
+#Singular function to create a resource link.
+def create_resource_link(created_resources, resource_link_entity, preview_mode = False):
+    # template scaffolding
     reference_json_block = {
         "reference" : "$value"
     }
-
+    #Special reference handling blocks, in the form of (origin_resource, destination_resource, reference_path)
     arrayType_references = [
         ('diagnosticreport', 'specimen', 'specimen'),
         ('diagnosticreport', 'practitioner', 'performer'),
@@ -143,35 +187,36 @@ def create_resource_links(created_resources, resource_link_entites):
         ('diagnosticreport', 'observation', 'result'),
         ('diagnosticreport', 'imagingStudy', 'imagingStudy'),
     ]
-    #TODO: Build resource links
-    print("Building resource links")
-    for resource_link_entity in resource_link_entites:
-        try:
-            origin_resource = created_resources[resource_link_entity.origin_resource]
-        except KeyError:
-            print(f"WARNING: In ResourceLinks tab, found a Origin Resource of : {resource_link_entity.origin_resource}  but no such entity found in PatientData")
-            continue
-        try:
-            destination_resource = created_resources[resource_link_entity.destination_resource]
-        except KeyError:
-            print(f"WARNING: In ResourceLinks tab, found a Desitnation Resource  of : {resource_link_entity.destination_resource}  but no such entity found in PatientData")
-            continue
-        destination_resource_type = created_resources[resource_link_entity.destination_resource]['resourceType']
-        destination_resource_id = created_resources[resource_link_entity.destination_resource]['id']
-        link_tuple = (created_resources[resource_link_entity.origin_resource]['resourceType'].strip().lower(),
-                      created_resources[resource_link_entity.destination_resource]['resourceType'].strip().lower(),
-                      resource_link_entity.reference_path.strip().lower())
-        if link_tuple in arrayType_references:
-            if resource_link_entity.reference_path.strip().lower() not in origin_resource:
-                origin_resource[resource_link_entity.reference_path.strip().lower()] = []
-            new_reference = reference_json_block.copy()
-            new_reference['reference'] = destination_resource_type + "/" + destination_resource_id
-            origin_resource[resource_link_entity.reference_path.strip().lower()].append(new_reference)
-        else:
-            origin_resource[resource_link_entity.reference_path.strip().lower()] = reference_json_block.copy()
-            origin_resource[resource_link_entity.reference_path.strip().lower()]["reference"] = destination_resource_type + "/" + destination_resource_id
+    #Find the origin and destination resource from the link
+    try:
+        origin_resource = created_resources[resource_link_entity.origin_resource]
+    except KeyError:
+        print(f"WARNING: In ResourceLinks tab, found a Origin Resource of : {resource_link_entity.origin_resource}  but no such entity found in PatientData")
+        return
+    try:
+        destination_resource = created_resources[resource_link_entity.destination_resource]
+    except KeyError:
+        print(f"WARNING: In ResourceLinks tab, found a Desitnation Resource  of : {resource_link_entity.destination_resource}  but no such entity found in PatientData")
+        return
+    #Estable the value of the refence
+    if preview_mode:
+        reference_value = destination_resource['resourceType'] + "/" + resource_link_entity.destination_resource
+    else:
+        reference_value = destination_resource['resourceType'] + "/" + destination_resource['id']
+    link_tuple = (origin_resource['resourceType'].strip().lower(),
+                    destination_resource['resourceType'].strip().lower(),
+                    resource_link_entity.reference_path.strip().lower())
+    if link_tuple in arrayType_references:
+        if resource_link_entity.reference_path.strip().lower() not in origin_resource:
+            origin_resource[resource_link_entity.reference_path.strip().lower()] = []
+        new_reference = reference_json_block.copy()
+        new_reference['reference'] = reference_value
+        origin_resource[resource_link_entity.reference_path.strip().lower()].append(new_reference)
+    else:
+        origin_resource[resource_link_entity.reference_path.strip().lower()] = reference_json_block.copy()
+        origin_resource[resource_link_entity.reference_path.strip().lower()]["reference"] = reference_value
     return
-    
+
 def add_resource_to_transaction_bundle(root_bundle, fhir_resource):
     entry = {}
     entry['fullUrl'] = "urn:uuid:"+fhir_resource['id']
@@ -192,7 +237,7 @@ def add_resource_to_transaction_bundle(root_bundle, fhir_resource):
 # resource_definition: resource description model from import
 # entity_definition: specific field entry information for this function
 # value: Actual value to assign
-def create_structure_from_jsonpath(root_struct: Dict, json_path: str, resource_definition: ResourceDefinition, field_entry: FieldEntry, dataType: str, value: Any):
+def create_structure_from_jsonpath(root_struct: Dict, json_path: str, resource_definition: ResourceDefinition,  dataType: str, value: Any):
     #Get all dot notation components as seperate 
     if dataType is not None and dataType.strip().lower() == 'string':
         value = str(value)
@@ -202,10 +247,10 @@ def create_structure_from_jsonpath(root_struct: Dict, json_path: str, resource_d
         return root_struct
     #Start of top-level function which calls the enclosed recursive function
     parts = json_path.split('.')
-    return build_structure(root_struct, json_path, resource_definition, field_entry, parts, value, [])
+    return build_structure(root_struct, json_path, resource_definition, dataType, parts, value, [])
 
 # main recursive function to drill into the json structure, assign paths, and create structure where needed
-def build_structure(current_struct, json_path, resource_definition, entity_definition, parts, value, previous_parts):
+def build_structure(current_struct: Dict, json_path: str, resource_definition: ResourceDefinition, dataType: str, parts: List[str], value: Any, previous_parts: List[str]):
     if len(parts) == 0:
         return current_struct
     #Grab current part
@@ -213,11 +258,11 @@ def build_structure(current_struct, json_path, resource_definition, entity_defin
     #SPECIAL HANDLING CLAUSE
     matching_handler = next((handler for handler in special_values.custom_handlers if (json_path.startswith(handler) or json_path == handler)), None)
     if matching_handler is not None:
-        return special_values.custom_handlers[matching_handler].assign_value(json_path, resource_definition, entity_definition, current_struct, parts[-1], value)
+        return special_values.custom_handlers[matching_handler].assign_value(json_path, resource_definition, dataType,  current_struct, parts[-1], value)
     #Ignore dollar sign ($) and drill farther down
     if part == '$' or part == resource_definition.resource_type.strip():
         #Ignore the dollar sign and the resourcetype
-        return build_structure_recurse(current_struct, json_path, resource_definition, entity_definition, parts, value, previous_parts, part)
+        return build_structure_recurse(current_struct, json_path, resource_definition, dataType, parts, value, previous_parts, part)
     
     # If parts length is one then this is the final key to access and pair
     if len(parts) == 1:
@@ -240,7 +285,7 @@ def build_structure(current_struct, json_path, resource_definition, entity_defin
                 if part + 1 > len(current_struct):
                     current_struct.extend({} for x in range (part + 1 - len(current_struct)))
         #Actual assigning to the path
-        fhir_formatting.assign_value(current_struct, part, value, entity_definition.value_type)
+        fhir_formatting.assign_value(current_struct, part, value, dataType)
         return current_struct
     
     # If there is a simple qualifier with '['and ']'
@@ -262,7 +307,7 @@ def build_structure(current_struct, json_path, resource_definition, entity_defin
             if qualifier_as_number + 1 > len(current_struct):
                 current_struct.extend({} for x in range (qualifier_as_number + 1 - len(current_struct)))
             inner_struct = current_struct[qualifier_as_number]
-            inner_struct = build_structure_recurse(inner_struct, json_path, resource_definition, entity_definition, parts, value, previous_parts, part)
+            inner_struct = build_structure_recurse(inner_struct, json_path, resource_definition, dataType, parts, value, previous_parts, part)
             current_struct[qualifier_as_number] = inner_struct
             return current_struct
         # Create the key part in the structure
@@ -284,7 +329,7 @@ def build_structure(current_struct, json_path, resource_definition, entity_defin
                 inner_struct = {qualifier_key: qualifier_value}
                 current_struct[key_part].append(inner_struct)
             #Recurse into that innerstructure where the qualifier matched to continue the part traversal
-            inner_struct = build_structure_recurse(inner_struct, json_path, resource_definition, entity_definition, parts, value, previous_parts, part)
+            inner_struct = build_structure_recurse(inner_struct, json_path, resource_definition, dataType, parts, value, previous_parts, part)
             return current_struct
         #If there's no qualifier condition, but an index aka '[0]', '[1]' etc, then it's a simple accessor
         elif qualifier.isdigit():
@@ -294,19 +339,19 @@ def build_structure(current_struct, json_path, resource_definition, entity_defin
             if qualifier_as_number > len(current_struct):
                 current_struct[key_part].extend({} for x in range (qualifier_as_number - len(current_struct)))
             inner_struct = current_struct[key_part][qualifier_as_number]
-            inner_struct = build_structure_recurse(inner_struct, json_path, resource_definition, entity_definition, parts, value, previous_parts, part)
+            inner_struct = build_structure_recurse(inner_struct, json_path, resource_definition, dataType, parts, value, previous_parts, part)
             current_struct[key_part][qualifier_as_number] = inner_struct
             return current_struct
     #None qualifier accessor
     else:
         if(part not in current_struct):
             current_struct[part] = {}
-        inner_struct = build_structure_recurse(current_struct[part], json_path, resource_definition, entity_definition, parts, value, previous_parts, part)
+        inner_struct = build_structure_recurse(current_struct[part], json_path, resource_definition, dataType, parts, value, previous_parts, part)
         current_struct[part] = inner_struct
         return current_struct
     
 #Helper function to quickly recurse and return the next level of structure. Used by main recursive function
-def build_structure_recurse(current_struct, json_path, resource_definition, entity_definition, parts, value, previous_parts, part):
+def build_structure_recurse(current_struct, json_path, resource_definition, dataType, parts, value, previous_parts, part):
     previous_parts.append(part)
-    return_struct = build_structure(current_struct, json_path, resource_definition, entity_definition, parts[1:], value, previous_parts)
+    return_struct = build_structure(current_struct, json_path, resource_definition, dataType, parts[1:], value, previous_parts)
     return return_struct
\ No newline at end of file
